# Adapted from: https://github.com/ryan112358/private-pgm/blob/master/src/mbi/domain.py

from functools import reduce


class Domain:
    def __init__(self, attrs, shape):
        """Construct a Domain object

        :param attrs: a list or tuple of attribute names
        :param shape: a list or tuple of domain sizes for each attribute
        """
        assert len(attrs) == len(shape), "dimensions must be equal"
        self.attrs = tuple(attrs)
        self.shape = tuple(shape)
        self.config = dict(zip(attrs, shape))

    @staticmethod
    def fromdict(config):
        """Construct a Domain object from a dictionary of { attr : size } values"""
        return Domain(config.keys(), config.values())

    def project(self, attrs):
        """project the domain onto a subset of attributes

        :param attrs: the attributes to project onto
        :return: the projected Domain object
        """
        # return the projected domain
        if type(attrs) is str:
            attrs = [attrs]
        shape = tuple(self.config[a] for a in attrs)
        return Domain(attrs, shape)

    def marginalize(self, attrs):
        """marginalize out some attributes from the domain (opposite of project)

        :param attrs: the attributes to marginalize out
        :return: the marginalized Domain object
        """
        proj = [a for a in self.attrs if a not in attrs]
        return self.project(proj)

    def axes(self, attrs):
        """return the axes tuple for the given attributes

        :param attrs: the attributes
        :return: a tuple with the corresponding axes
        """
        return tuple(self.attrs.index(a) for a in attrs)

    def transpose(self, attrs):
        """reorder the attributes in the domain object"""
        return self.project(attrs)

    def invert(self, attrs):
        """returns the attributes in the domain not in the list"""
        return [a for a in self.attrs if a not in attrs]

    def merge(self, other):
        """merge this domain object with another

        :param other: another Domain object
        :return: a new domain object covering the full domain

        Example:
        >>> D1 = Domain(['a','b'], [10,20])
        >>> D2 = Domain(['b','c'], [20,30])
        >>> D1.merge(D2)
        Domain(['a','b','c'], [10,20,30])
        """
        extra = other.marginalize(self.attrs)
        return Domain(self.attrs + extra.attrs, self.shape + extra.shape)

    def contains(self, other):
        """determine if this domain contains another"""
        return set(other.attrs) <= set(self.attrs)

    def size(self, attrs=None):
        """return the total size of the domain"""
        if attrs is None:
            return reduce(lambda x, y: x * y, self.shape, 1)
        return self.project(attrs).size()

    def sort(self, how="size"):
        """return a new domain object, sorted by attribute size or attribute name"""
        if how == "size":
            attrs = sorted(self.attrs, key=self.size)
        elif how == "name":
            attrs = sorted(self.attrs)
        return self.project(attrs)

    def canonical(self, attrs):
        """return the canonical ordering of the attributes"""
        return tuple(a for a in self.attrs if a in attrs)

    def __contains__(self, attr):
        return attr in self.attrs

    def __getitem__(self, a):
        """return the size of an individual attribute
        :param a: the attribute
        """
        return self.config[a]

    def __iter__(self):
        """iterator for the attributes in the domain"""
        return self.attrs.__iter__()

    def __len__(self):
        return len(self.attrs)

    def __eq__(self, other):
        return self.attrs == other.attrs and self.shape == other.shape

    def __repr__(self):
        inner = ", ".join(["%s: %d" % x for x in zip(self.attrs, self.shape)])
        return "Domain(%s)" % inner

    def __str__(self):
        return self.__repr__()



from typing import Tuple, List
import numpy as np

from dpmm.models.base.mbi.dataset import Dataset


def meas_size(shape: Tuple):
    return np.prod(shape) / (8 * (1028 ** 2))


def clique_size(data: Dataset, clique: Tuple):
    return meas_size(data.project(clique).domain.shape)


def model_size(data: Dataset, cliques: List[Tuple]):
    return sum([clique_size(data, clique) for clique in cliques])



# This file is part of the Ektelo framework
# For licensing terms please see: https://github.com/ektelo/ektelo
#
# Copyright 2019-2021, Tumult Labs Inc.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import LinearOperator

# from functools import reduce
# import math


class EkteloMatrix(LinearOperator):
    """
    An EkteloMatrix is a linear transformation that can compute matrix-vector products
    """

    # must implement: _matmat, _transpose, matrix
    # can  implement: gram, sensitivity, sum, dense_matrix, sparse_matrix, __abs__

    def __init__(self, matrix):
        """Instantiate an EkteloMatrix from an explicitly represented backing matrix

        :param matrix: a 2d numpy array or a scipy sparse matrix
        """
        self.matrix = matrix
        self.dtype = matrix.dtype
        self.shape = matrix.shape

    def _transpose(self):
        return EkteloMatrix(self.matrix.T)

    def _matmat(self, V):
        """
        Matrix multiplication of a m x n matrix Q

        :param V: a n x p numpy array
        :return Q*V: a m x p numpy aray
        """
        return self.matrix @ V

    def gram(self):
        """
        Compute the Gram matrix of the given matrix.
        For a matrix Q, the gram matrix is defined as Q^T Q
        """
        return self.T @ self  # works for subclasses too

    def sensitivity(self):
        # note: this works because np.abs calls self.__abs__
        return np.max(np.abs(self).sum(axis=0))

    def sum(self, axis=None):
        # this implementation works for all subclasses too
        # (as long as they define _matmat and _transpose)
        if axis == 0:
            return self.T.dot(np.ones(self.shape[0]))
        ans = self.dot(np.ones(self.shape[1]))
        return ans if axis == 1 else np.sum(ans)

    def inv(self):
        return EkteloMatrix(np.linalg.inv(self.dense_matrix()))

    def pinv(self):
        return EkteloMatrix(np.linalg.pinv(self.dense_matrix()))

    def trace(self):
        return self.diag().sum()

    def diag(self):
        return np.diag(self.dense_matrix())

    def _adjoint(self):
        return self._transpose()

    def __mul__(self, other):
        if np.isscalar(other):
            return Weighted(self, other)
        if type(other) == np.ndarray:
            return self.dot(other)
        if isinstance(other, EkteloMatrix):
            return Product(self, other)
            # note: this expects both matrix types to be compatible (e.g., sparse and sparse)
            # todo: make it work for different backing representations
        else:
            raise TypeError(
                "incompatible type %s for multiplication with EkteloMatrix"
                % type(other)
            )

    def __add__(self, other):
        if np.isscalar(other):
            other = Weighted(Ones(self.shape), other)
        return Sum([self, other])

    def __sub__(self, other):
        return self + -1 * other

    def __rmul__(self, other):
        if np.isscalar(other):
            return Weighted(self, other)
        return NotImplemented

    def __getitem__(self, key):
        """
        return a given row from the matrix

        :param key: the index of the row to return
        :return: a 1xN EkteloMatrix
        """
        # row indexing, subclasses may provide more efficient implementation
        m, n = self.shape
        v = np.zeros(m)
        v[key] = 1.0
        return EkteloMatrix(self.T.dot(v).reshape(1, n))

    def dense_matrix(self):
        """
        return the dense representation of this matrix, as a 2D numpy array
        """
        if sparse.issparse(self.matrix):
            return self.matrix.toarray()
        return self.matrix

    def sparse_matrix(self):
        """
        return the sparse representation of this matrix, as a scipy matrix
        """
        if sparse.issparse(self.matrix):
            return self.matrix
        return sparse.csr_matrix(self.matrix)

    @property
    def ndim(self):
        # todo: deprecate if possible
        return 2

    def __abs__(self):
        return EkteloMatrix(self.matrix.__abs__())

    def __sqr__(self):
        if sparse.issparse(self.matrix):
            return EkteloMatrix(self.matrix.power(2))
        return EkteloMatrix(self.matrix**2)

    def l1_sensitivity(self):
        return self.__abs__().sum(axis=0).max()

    def l2_sensitivity(self):
        return np.sqrt(self.__sqr__().sum(axis=0).max())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __hash__(self):
        return hash(repr(self))


class Identity(EkteloMatrix):
    def __init__(self, n, dtype=np.float64):
        self.n = n
        self.shape = (n, n)
        self.dtype = dtype

    def _matmat(self, V):
        return V

    def _transpose(self):
        return self

    @property
    def matrix(self):
        return sparse.eye(self.n, dtype=self.dtype)

    def __mul__(self, other):
        assert other.shape[0] == self.n, "dimension mismatch"
        return other

    def inv(self):
        return self

    def pinv(self):
        return self

    def trace(self):
        return self.n

    def __abs__(self):
        return self

    def __sqr__(self):
        return self


class Ones(EkteloMatrix):
    """A m x n matrix of all ones"""

    def __init__(self, m, n, dtype=np.float64):
        self.m = m
        self.n = n
        self.shape = (m, n)
        self.dtype = dtype

    def _matmat(self, V):
        ans = V.sum(axis=0, keepdims=True)
        return np.repeat(ans, self.m, axis=0)

    def _transpose(self):
        return Ones(self.n, self.m, self.dtype)

    def gram(self):
        return self.m * Ones(self.n, self.n, self.dtype)

    def pinv(self):
        c = 1.0 / (self.m * self.n)
        return c * Ones(self.n, self.m, self.dtype)

    def trace(self):
        assert self.n == self.m, "matrix is not square"
        return self.n

    @property
    def matrix(self):
        return np.ones(self.shape, dtype=self.dtype)

    def __abs__(self):
        return self

    def __sqr__(self):
        return self


class Weighted(EkteloMatrix):
    """Class for multiplication by a constant"""

    def __init__(self, base, weight):
        if isinstance(base, Weighted):
            weight *= base.weight
            base = base.base
        self.base = base
        self.weight = weight
        self.shape = base.shape
        self.dtype = base.dtype

    def _matmat(self, V):
        return self.weight * self.base.dot(V)

    def __mul__(self, other):
        if isinstance(other, EkteloMatrix):
            return Weighted(self.base @ other, self.weight)
        return EkteloMatrix.__mul__(self, other)

    def _transpose(self):
        return Weighted(self.base.T, self.weight)

    def gram(self):
        return Weighted(self.base.gram(), self.weight**2)

    def pinv(self):
        return Weighted(self.base.pinv(), 1.0 / self.weight)

    def inv(self):
        return Weighted(self.base.inv(), 1.0 / self.weight)

    def trace(self):
        return self.weight * self.base.trace()

    def __abs__(self):
        return Weighted(self.base.__abs__(), np.abs(self.weight))

    def __sqr__(self):
        return Weighted(self.base.__sqr__(), self.weight**2)

    def sensitivity(self):
        return self.weight * self.base.sensitivity()

    @property
    def matrix(self):
        return self.weight * self.base.matrix


class Sum(EkteloMatrix):
    """Class for the Sum of matrices"""

    def __init__(self, matrices):
        # all must have same shape
        self.matrices = matrices
        self.shape = matrices[0].shape
        self.dtype = np.result_type(*[Q.dtype for Q in matrices])

    def _matmat(self, V):
        return sum(Q.dot(V) for Q in self.matrices)

    def _transpose(self):
        return Sum([Q.T for Q in self.matrices])

    def __mul__(self, other):
        if isinstance(other, EkteloMatrix):
            return Sum(
                [Q @ other for Q in self.matrices]
            )  # should use others rmul though
        return EkteloMatrix.__mul__(self, other)

    def diag(self):
        return sum(Q.diag() for Q in self.matrices)

    def trace(self):
        return sum(Q.trace() for Q in self.matrices)

    @property
    def matrix(self):
        if _any_sparse(self.matrices):
            return sum(Q.sparse_matrix() for Q in self.matrices)
        return sum(Q.dense_matrix() for Q in self.matrices)


# class BlockDiag(EkteloMatrix):
#     def __init__(self, matrices):
#         self.matrices = matrices
#         rows = sum(Q.shape[0] for Q in matrices)
#         cols = sum(Q.shape[1] for Q in matrices)
#         self.shape = (rows, cols)
#         self.dtype = np.result_type(*[Q.dtype for Q in matrices])
#
#     # TODO: implement _matmat
#
#     def diag(self):
#         return np.concatenate([Q.diag() for Q in self.matrices])
#
#     @property
#     def matrix(self):
#         return sparse.block_diag([Q.matrix for Q in self.matrices], format='csr')
#
#
# class VStack(EkteloMatrix):
#     def __init__(self, matrices):
#         m = sum(Q.shape[0] for Q in matrices)
#         n = matrices[0].shape[1]
#         assert all(Q.shape[1] == n for Q in matrices), 'dimension mismatch'
#         self.shape = (m, n)
#         self.matrices = matrices
#         self.dtype = np.result_type(*[Q.dtype for Q in matrices])
#
#     def _matmat(self, V):
#         return np.vstack([Q.dot(V) for Q in self.matrices])
#
#     def _transpose(self):
#         return HStack([Q.T for Q in self.matrices])
#
#     def __mul__(self, other):
#         if isinstance(other, EkteloMatrix):
#             return VStack([Q @ other for Q in self.matrices])  # should use others rmul though
#         return EkteloMatrix.__mul__(self, other)
#
#     def gram(self):
#         return Sum([Q.gram() for Q in self.matrices])
#
#     @property
#     def matrix(self):
#         if _any_sparse(self.matrices):
#             return self.sparse_matrix()
#         return self.dense_matrix()
#
#     def dense_matrix(self):
#         return np.vstack([Q.dense_matrix() for Q in self.matrices])
#
#     def sparse_matrix(self):
#         return sparse.vstack([Q.sparse_matrix() for Q in self.matrices])
#
#     def __abs__(self):
#         return VStack([Q.__abs__() for Q in self.matrices])
#
#     def __sqr__(self):
#         return VStack([Q.__sqr__() for Q in self.matrices])
#
#
# class HStack(EkteloMatrix):
#     def __init__(self, matrices):
#         # all matrices must have same number of rows
#         cols = [Q.shape[1] for Q in matrices]
#         m = matrices[0].shape[0]
#         n = sum(cols)
#         assert all(Q.shape[0] == m for Q in matrices), 'dimension mismatch'
#         self.shape = (m, n)
#         self.matrices = matrices
#         self.dtype = np.result_type(*[Q.dtype for Q in matrices])
#         self.split = np.cumsum(cols)[:-1]
#
#     def _matmat(self, V):
#         vs = np.split(V, self.split)
#         ans = np.zeros((self.shape[0], V.shape[1]), dtype=self.dtype)
#         for Q, z in zip(self.matrices, vs):
#             ans += Q.dot(z)
#         return ans
#
#     def _transpose(self):
#         return VStack([Q.T for Q in self.matrices])
#
#     @property
#     def matrix(self):
#         if _any_sparse(self.matrices):
#             return self.sparse_matrix()
#         return self.dense_matrix()
#
#     def dense_matrix(self):
#         return np.hstack([Q.dense_matrix() for Q in self.matrices])
#
#     def sparse_matrix(self):
#         return sparse.hstack([Q.sparse_matrix() for Q in self.matrices])
#
#     def __mul__(self, other):
#         if isinstance(other, VStack):
#             # and shapes match...
#             return Sum([A @ B for A, B in zip(self.matrices, other.matrices)])
#         return EkteloMatrix.__mul__(self, other)
#
#     def __rmul__(self, other):
#         if isinstance(other, EkteloMatrix):
#             return HStack([other @ Q for Q in self.matrices])
#         return EkteloMatrix.__mul__(self, other)
#
#     def __abs__(self):
#         return HStack([Q.__abs__() for Q in self.matrices])
#
#     def __sqr__(self):
#         return HStack([Q.__sqr__() for Q in self.matrices])
#
#
# class Kronecker(EkteloMatrix):
#     def __init__(self, matrices):
#         self.matrices = matrices
#         self.shape = tuple(np.prod([Q.shape for Q in matrices], axis=0))
#         self.dtype = np.result_type(*[Q.dtype for Q in matrices])
#
#     def _matmat(self, V):
#         X = V.T
#         for Q in self.matrices[::-1]:
#             m, n = Q.shape
#             X = Q.dot(X.reshape(-1, n).T)
#         return X.reshape(self.shape[0], -1)
#
#     def _transpose(self):
#         return Kronecker([Q.T for Q in self.matrices])
#
#     def gram(self):
#         return Kronecker([Q.gram() for Q in self.matrices])
#
#     @property
#     def matrix(self):
#         if _any_sparse(self.matrices):
#             return self.sparse_matrix()
#         return self.dense_matrix()
#
#     def dense_matrix(self):
#         return reduce(np.kron, [Q.dense_matrix() for Q in self.matrices])
#
#     def sparse_matrix(self):
#         return reduce(sparse.kron, [Q.sparse_matrix() for Q in self.matrices])
#
#     def sensitivity(self):
#         return np.prod([Q.sensitivity() for Q in self.matrices])
#
#     def inv(self):
#         return Kronecker([Q.inv() for Q in self.matrices])
#
#     def pinv(self):
#         return Kronecker([Q.pinv() for Q in self.matrices])
#
#     def diag(self):
#         return reduce(np.kron, [Q.diag() for Q in self.matrices])
#
#     def trace(self):
#         return np.prod([Q.trace() for Q in self.matrices])
#
#     def __mul__(self, other):
#         # perform the multiplication in the implicit representation if possible
#         if isinstance(other, Kronecker):
#             return Kronecker([A @ B for A, B in zip(self.matrices, other.matrices)])
#         elif isinstance(other, HStack):
#             return other.__rmul__(self)
#         return EkteloMatrix.__mul__(self, other)
#
#     def __abs__(self):
#         return Kronecker([Q.__abs__() for Q in self.matrices])
#
#     def __sqr__(self):
#         return Kronecker([Q.__sqr__() for Q in self.matrices])
#
#
# class Haar(EkteloMatrix):
#     """
#     The Haar wavelet is a square matrix of size n x n where n is a power of 2
#     """
#     def __init__(self, n, dtype=np.float64):
#         self.n = n
#         self.k = int(math.log(n, 2))
#         assert 2**self.k == n, 'n must be a power of 2'
#         self.shape = (n, n)
#         self.dtype = dtype
#
#     def _matmat(self, X):
#         y = X.copy()
#         n = self.n
#         for _ in range(self.k):
#             y[:n] = np.vstack([y[:n][0::2] + y[:n][1::2], y[:n][0::2] - y[:n][1::2]])
#             n = n // 2
#         return y
#
#     def _rmatvec(self, y):
#         # can implement this instead of _transpose
#         x = y.copy()
#         m = 1
#         for _ in range(self.k):
#             n = 2 * m
#             # be careful here, don't separate into two calls
#             x[0:n:2], x[1:n:2] = x[:m] + x[m:n], x[:m] - x[m:n]
#             m *= 2
#         return x
#
#     def _transpose(self):
#         return LinearOperator._adjoint(self)
#
#     def sensitivity(self):
#         return self.k + 1.0
#
#     @property
#     def matrix(self):
#         H = sparse.eye(1, format='csr')
#         for m in [2**c for c in range(self.k)]:
#             I = sparse.eye(m, format='csr')
#             A = sparse.kron(H, [1, 1], format='csr')
#             B = sparse.kron(I, [1, -1], format='csr')
#             H = sparse.vstack([A, B], format='csr')
#         return H


class Product(EkteloMatrix):
    def __init__(self, A, B):
        assert A.shape[1] == B.shape[0]
        self._A = A
        self._B = B
        self.shape = (A.shape[0], B.shape[1])
        self.dtype = np.result_type(A.dtype, B.dtype)

    def _matmat(self, X):
        return self._A.dot(self._B.dot(X))

    def _transpose(self):
        return Product(self._B.T, self._A.T)

    @property
    def matrix(self):
        return self._A.matrix @ self._B.matrix

    def gram(self):
        return Product(self.T, self)

    def inv(self):
        return Product(self._B.inv(), self._A.inv())


def _any_sparse(matrices):
    return any(sparse.issparse(Q.matrix) for Q in matrices)



import json
from pathlib import Path

from dpmm.models.aim import AIMGM
from dpmm.models.base.base import GenerativeModel
from dpmm.models.mst import MSTGM
from dpmm.models.priv_bayes import PrivBayesGM

MODELS = [PrivBayesGM, AIMGM, MSTGM]
MODEL_DICT = {MODEL.name: MODEL for MODEL in MODELS}


def load_model(path: Path) -> GenerativeModel:
    with (path / "model_type.json").open("r") as fr:
        model_type = json.load(fr)

    return MODEL_DICT[model_type].load(path)



# A generative model training algorithm based on
# "PrivBayes: Private Data Release via Bayesian Networks" 
# by Jun Zhang1 Graham Cormode2 Cecilia M. Procopiuc3 Divesh Srivastava3 Xiaokui Xiao1
# combined with the pgm framework by Ryan McKenna, Gerome Miklau, Daniel Sheldon
# code inspired from : https://github.com/ryan112358/private-pgm/blob/1da21c8b38149b05f1385b8e54116568b700b4fa/mechanisms/mst.py


from itertools import combinations
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, Self, Tuple

# import argparse
import joblib
import numpy as np
from numpy.random import RandomState
from scipy.special import logsumexp

from dpmm.models.base.graphical import GraphicalGenerativeModel
from dpmm.models.base.mbi import Dataset, GraphicalModel
from dpmm.models.base.mechanisms import cdp_rho
from dpmm.models.base.memory import clique_size
from dpmm.utils import to_path
from dpmm.models.base.mechanisms import Mechanism
from warnings import warn

"""
This is a generalization of the winning mechanism from the
2018 NIST Differential Privacy Synthetic Data Competition.

Unlike the original implementation, this one can work for any discrete dataset,
and does not rely on public provisional data for measurement selection.
"""


class MI:
    """
    Mutual Information (MI) computation class.

    :param data: The dataset.
    :type data: Dataset
    """

    def __init__(self, data: Dataset):
        self.cache = {}
        self.data = data

    def __call__(self, parents: Tuple[str], child: str) -> float:
        """
        Compute the mutual information between a child and its parents.

        :param parents: Tuple of parent attributes.
        :type parents: Tuple[str]
        :param child: Child attribute.
        :type child: str
        :return: Mutual information value.
        :rtype: float
        """
        parents = tuple(parents)
        cache_key = (child,) + parents

        if cache_key in self.cache:
            return self.cache[cache_key]

        joint_counts = self.data.project(list(parents + (child,))).datavector(
            flatten=False
        )

        joint_counts = np.nan_to_num(joint_counts, nan=0.0)
        normaliser = joint_counts.sum(axis=None)
        if normaliser == 0:
            normaliser = 1

        if child not in self.cache:
            p_1 = joint_counts.sum(axis=tuple(np.arange(len(parents)))) / normaliser
            self.cache[child] = e_1 = self.entropy(p_1)
        else:
            e_1 = self.cache[child]

        if parents not in self.cache:
            p_2 = joint_counts.sum(axis=-1) / normaliser
            self.cache[parents] = e_2 = self.entropy(p_2)
        else:
            e_2 = self.cache[parents]

        p_1_2 = joint_counts / normaliser
        e_1_2 = self.entropy(p_1_2)
        if e_1 == 0 or e_2 == 0:
            value = 0
        else:
            value = (e_1 + e_2 - e_1_2) / np.sqrt(e_1 * e_2)
        self.cache[cache_key] = value
        return value

    def entropy(self, p: np.ndarray) -> float:
        """
        Calculate the entropy of a probability distribution.

        :param p: Probability distribution.
        :type p: np.ndarray
        :return: Entropy value.
        :rtype: float
        """
        pp = p[p > 0]
        return -np.sum(pp * np.log2(pp), axis=None)


class ExponentialMechanism:
    """
    Exponential Mechanism for differential privacy.

    :param data: The dataset.
    :type data: Dataset
    :param epsilon: Privacy budget.
    :type epsilon: float
    :param normalise: Whether to normalise the output probabilities.
    :type normalise: bool
    """

    def __init__(self, data: Dataset, epsilon: float, normalise: bool = True):
        self.epsilon = epsilon
        self.delta = self.compute_delta(data)
        self.normalise = normalise

    def compute_delta(self, data: Dataset) -> float:
        """
        Compute the sensitivity delta for the dataset.

        :param data: The dataset.
        :type data: Dataset
        :return: Sensitivity delta.
        :rtype: float
        """
        n_rows = data.df.shape[0]
        n_nodes = len(data.domain.attrs)
        a = (2 / n_rows) * np.log((n_rows + 1) / 2)
        b = (1 - 1 / n_rows) * np.log(1 + 2 / (n_rows - 1))
        sensitivity = a + b
        eps = self.epsilon / n_nodes
        delta = sensitivity / eps
        return delta

    def __call__(self, input_vector: np.ndarray) -> np.ndarray:
        """
        Apply the exponential mechanism to an input vector.

        :param input_vector: Input vector.
        :type input_vector: np.ndarray
        :return: Output probabilities.
        :rtype: np.ndarray
        """
        with_delta = np.array(input_vector) / (2 * self.delta)

        if self.normalise:
            log_normalised = with_delta - logsumexp(with_delta)
            normalised = np.exp(log_normalised)
            normalised /= normalised.sum()
            return normalised

        return np.exp(with_delta)


class PrivBayes(Mechanism):
    """
    PrivBayes is a differentially private generative model that uses Bayesian networks
    to approximate the joint distribution of the data. It selects dependencies between
    attributes based on mutual information and relies on
    the exponential and laplace mechanism to ensure differential privacy.

    Ref: https://dl.acm.org/doi/pdf/10.1145/3134428

    :param epsilon: Privacy budget.
    :type epsilon: float
    :param delta: Privacy parameter.
    :type delta: float, optional
    :param degree: Maximum degree of the Bayesian network. The size of the largest clique.
    :type degree: int
    :param n_iters: Number of iterations for fitting.
    :type n_iters: int
    :param max_candidates: Maximum number of candidates for selection.
    :type max_candidates: int
    :param n_jobs: Number of parallel jobs.
    :type n_jobs: int
    :param prng: Random state for reproducibility.
    :type prng: RandomState, optional
    :param max_model_size: Maximum size of the model.
    :type max_model_size: int, optional
    :param compress: Whether to compress the data.
    :type compress: bool
    :param domain: The domain of the data.
    :type domain: Domain, optional
    :param structural_zeros: Structural zeros in the data.
    :type structural_zeros: Dict, optional
    """

    def __init__(
        self,
        epsilon=1,
        delta=None,
        degree=2,
        n_iters: int = 5000,
        max_candidates: int = 20,
        n_jobs: int = -1,
        prng: RandomState = None,
        max_model_size: int = None,
        compress=False,
        domain=None,
        structural_zeros: Dict = None,
    ):
        super().__init__(
            epsilon=epsilon,
            delta=delta,
            prng=prng,
            max_model_size=max_model_size,
            compress=compress,
            domain=domain,
            structural_zeros=structural_zeros,
            n_jobs=n_jobs,
        )

        self.rho = cdp_rho(self.epsilon / 2, self.delta)
        self.sigma = np.sqrt(1 / (2 * self.rho))

        self.degree = degree
        self.n_iters = n_iters
        self.max_candidates = max_candidates

    def select(self, data: Dataset, public=False) -> list:
        """
        Select cliques based on mutual information.

        :param data: The dataset.
        :type data: Dataset
        :param public: Whether to use public data for selection.
        :type public: bool
        :return: List of selected cliques.
        :rtype: list
        """
        remaining = set(data.domain.attrs)
        used = set()
        cliques = []

        attrs = data.domain.attrs
        if self.max_model_size is not None:
            size_per_node = self.max_model_size / len(attrs)
            attrs = [
                attr
                for attr in attrs
                if clique_size(data, clique=(attr)) <= size_per_node
            ]
            if len(attrs) == 0:
                attrs = data.domain.attrs
        root = self.prng.choice(attrs)
        used.add(root)
        remaining.remove(root)
        cliques.append((root,))

        mi = MI(data=data)
        if not public:
            exp_mechanism = ExponentialMechanism(
                data=data, epsilon=self.epsilon / 2, normalise=True
            )

        while len(remaining) > 0:
            degree = min(len(used), self.degree)
            candidates = []
            combs = list(combinations(used, degree))
            for col in remaining:
                col_candidates = combs
                if self.max_model_size is not None:
                    col_candidates = [
                        candidate
                        for candidate in col_candidates
                        if clique_size(data, clique=((candidate) + (col,)))
                        <= size_per_node
                    ]
                if len(col_candidates) == 0:
                    min_size = min([
                        clique_size(data, clique=((candidate) + (col,)))
                        for candidate in col_candidates
                    ])
                    warn(f"No combination respects the clique size limitation, will be skipped. Min Size : {min_size} | expected : {size_per_node}")
                    col_candidates = list(combs)

                if self.max_candidates is not None:
                    if self.max_candidates < len(col_candidates):
                        selected_idx = list(
                            self.prng.choice(
                                len(col_candidates),
                                size=self.max_candidates,
                                replace=False,
                            )
                        )
                        col_candidates = [col_candidates[idx] for idx in selected_idx]

                candidates += [(candidate, col) for candidate in col_candidates]

            if self.n_jobs > 1:
                mi_scores = np.array(
                    [mi_score for mi_score in Pool(self.n_jobs).starmap(mi, candidates)]
                )
            else:
                mi_scores = np.array(
                    [mi(*candidate) for candidate in candidates]
                )

            if public:
                selected = np.argmax(mi_scores)
            else:
                p = exp_mechanism(mi_scores)
                selected = self.prng.choice(len(candidates), p=p, size=1)[0]

            selected_candidate, selected_col = candidates[selected]
            selected_clique = selected_candidate + (selected_col,)

            cliques.append(selected_clique)
            used.add(selected_col)
            remaining.remove(selected_col)

        return cliques

    def _fit(self, data: Dataset, public=False) -> Tuple[Dataset, np.ndarray]:
        """
        Fit the PrivBayes model to the data.

        :param data: The dataset.
        :type data: Dataset
        :param public: Whether to use public data for fitting.
        :type public: bool
        :return: Transformed data and measures.
        :rtype: Tuple[Dataset, np.ndarray]
        """
        self.cliques = self.select(data=data, public=public)
        measures = self.measure(data, public=public, flatten=not self.compress)

        if self.compress:
            measures = self.compressor.fit(measures, flatten=True)
            data = self.compressor.transform(data)

        return data, measures

    def store(self, path: Path) -> None:
        """
        Store the PrivBayes model state to a file.

        :param path: The path to store the state.
        :type path: Path
        """
        joblib.dump(
            {
                "epsilon": self.epsilon,
                "delta": self.delta,
                "n_iters": self.n_iters,
                "cliques": self.cliques,
                "compress": self._compress,
                "compressor": self.compressor,
                "degree": self.degree,
                "max_candidates": self.max_candidates,
                "domain": self._domain,
                "max_model_size": self.max_model_size,
                "model_size": self.model_size,
                "fit_state": self.fit_state,
            },
            path / "state.joblib",
        )
        if self.model is not None:
            self.model.save(path / "estimator.pickle")

    @classmethod
    def load(cls, path: Path) -> Self:
        """
        Load the PrivBayes model state from a file.

        :param path: The path to load the state.
        :type path: Path
        :return: The loaded PrivBayes model.
        :rtype: PrivBayes
        """
        state = joblib.load(path / "state.joblib")
        obj = cls(
            epsilon=state["epsilon"],
            delta=state["delta"],
            n_iters=state["n_iters"],
            degree=state["degree"],
            max_candidates=state["max_candidates"],
            compress=state["compress"],
            domain=state["domain"],
            max_model_size=state["max_model_size"],
        )

        obj.cliques = state["cliques"]
        obj.compressor = state["compressor"]
        obj.model_size = state["model_size"]
        obj.fit_state = state["fit_state"]

        model_path = path / "estimator.pickle"
        if model_path.exists():
            obj.model = GraphicalModel.load(model_path)
        return obj


class PrivBayesGM(GraphicalGenerativeModel):
    """
    PrivBayesGM is a wrapper for the PrivBayes generative model.

    :param epsilon: Privacy budget.
    :type epsilon: float
    :param delta: Privacy parameter.
    :type delta: float
    :param degree: Maximum degree of the Bayesian network.
    :type degree: int
    :param n_iters: Number of iterations for fitting.
    :type n_iters: int
    :param compress: Whether to compress the data.
    :type compress: bool
    :param max_model_size: Maximum size of the model.
    :type max_model_size: int, optional
    :param domain: The domain of the data.
    :type domain: Domain, optional
    :param random_state: Random state for reproducibility.
    :type random_state: RandomState, optional
    """

    name = "priv-bayes"
    mechansim = PrivBayes

    def __init__(
        self,
        epsilon=1,
        delta=1e-5,
        degree=2,
        n_iters: int = 5000,
        compress=True,
        max_model_size: int = None,
        domain=None,
        random_state: RandomState = None,
        n_jobs: int = -1,
    ):
        super().__init__(domain=domain, random_state=random_state)

        self.epsilon = epsilon
        self.delta = delta

        self.generator = self.mechansim(
            epsilon=self.epsilon,
            delta=self.delta,
            n_iters=n_iters,
            degree=degree,
            compress=compress,
            domain=domain,
            prng=random_state,
            max_model_size=max_model_size,
            n_jobs=n_jobs,
        )

    def set_random_state(self, random_state: RandomState) -> None:
        """
        Set the random state for reproducibility.

        :param random_state: Random state.
        :type random_state: RandomState
        """
        super().set_random_state(random_state)
        self.generator.set_random_state(random_state)

    def set_domain(self, domain: Dict) -> None:
        """
        Set the domain of the data.

        :param domain: The domain.
        :type domain: Dict
        """
        super().set_domain(domain=domain)
        self.generator.set_domain(domain=domain)

    @to_path
    def store(self, path: Path) -> None:
        """
        Store the PrivBayesGM model state to a file.

        :param path: The path to store the state.
        :type path: Path
        """
        super().store(path)
        self.generator.store(path)

    @classmethod
    @to_path
    def load(cls, path: Path) -> Self:
        """
        Load the PrivBayesGM model state from a file.

        :param path: The path to load the state.
        :type path: Path
        :return: The loaded PrivBayesGM model.
        :rtype: PrivBayesGM
        """
        generator = PrivBayes.load(path)
        obj = cls(epsilon=generator.epsilon, delta=generator.delta)
        del obj.generator
        obj.generator = generator

        return obj



# Adapted from: https://github.com/ryan112358/private-pgm/blob/master/src/mbi/factor.py

import numpy as np
from scipy.special import logsumexp


class Factor:
    def __init__(self, domain, values, prng: np.random.RandomState = None):
        """Initialize a factor over the given domain

        :param domain: the domain of the factor
        :param values: the ndarray of factor values (for each element of the domain)

        Note: values may be a flattened 1d array or a ndarray with same shape as domain
        """
        assert domain.size() == values.size, "domain size does not match values size"
        assert (
            values.ndim == 1 or values.shape == domain.shape
        ), "invalid shape for values array"
        self.domain = domain
        self.values = values.reshape(domain.shape)
        self._prng = prng

    def set_random_state(self, prng: np.random.RandomState):
        self._prng = prng

    @property
    def prng(self):
        if self._prng is None:
            self._prng = np.random.RandomState()
        return self._prng

    @staticmethod
    def zeros(domain):
        return Factor(domain, np.zeros(domain.shape))

    @staticmethod
    def ones(domain):
        return Factor(domain, np.ones(domain.shape))

    @staticmethod
    def random(domain, prng: np.random.RandomState = np.random):
        return Factor(domain, prng.rand(*domain.shape))

    @staticmethod
    def uniform(domain):
        return Factor.ones(domain) / domain.size()

    @staticmethod
    def active(domain, structural_zeros):
        """create a factor that is 0 everywhere except in positions present in
            'structural_zeros', where it is -infinity

        :param: domain: the domain of this factor
        :param: structural_zeros: a list of values that are not possible
        """
        idx = tuple(np.array(structural_zeros).T)
        vals = np.zeros(domain.shape)
        vals[idx] = -np.inf
        return Factor(domain, vals)

    def expand(self, domain):
        assert domain.contains(
            self.domain
        ), "expanded domain must contain current domain"
        dims = len(domain) - len(self.domain)
        values = self.values.reshape(self.domain.shape + tuple([1] * dims))
        ax = domain.axes(self.domain.attrs)
        values = np.moveaxis(values, range(len(ax)), ax)
        values = np.broadcast_to(values, domain.shape)
        return Factor(domain, values)

    def transpose(self, attrs):
        assert set(attrs) == set(
            self.domain.attrs
        ), "attrs must be same as domain attributes"
        newdom = self.domain.project(attrs)
        ax = newdom.axes(self.domain.attrs)
        values = np.moveaxis(self.values, range(len(ax)), ax)
        return Factor(newdom, values)

    def project(self, attrs, agg="sum"):
        """
        project the factor onto a list of attributes (in order)
        using either sum or logsumexp to aggregate along other attributes
        """
        assert agg in ["sum", "logsumexp"], "agg must be sum or logsumexp"
        marginalized = self.domain.marginalize(attrs)
        if agg == "sum":
            ans = self.sum(marginalized.attrs)
        elif agg == "logsumexp":
            ans = self.logsumexp(marginalized.attrs)
        return ans.transpose(attrs)

    def sum(self, attrs=None):
        if attrs is None:
            return np.sum(self.values)
        axes = self.domain.axes(attrs)
        values = np.sum(self.values, axis=axes)
        newdom = self.domain.marginalize(attrs)
        return Factor(newdom, values)

    def logsumexp(self, attrs=None):
        if attrs is None:
            return logsumexp(self.values)
        axes = self.domain.axes(attrs)
        values = logsumexp(self.values, axis=axes)
        newdom = self.domain.marginalize(attrs)
        return Factor(newdom, values)

    def logaddexp(self, other):
        newdom = self.domain.merge(other.domain)
        factor1 = self.expand(newdom)
        factor2 = self.expand(newdom)
        return Factor(newdom, np.logaddexp(factor1.values, factor2.values))

    def max(self, attrs=None):
        if attrs is None:
            return self.values.max()
        axes = self.domain.axes(attrs)
        values = np.max(self.values, axis=axes)
        newdom = self.domain.marginalize(attrs)
        return Factor(newdom, values)

    def condition(self, evidence):
        """evidence is a dictionary where
        keys are attributes, and
        values are elements of the domain for that attribute"""
        slices = [evidence[a] if a in evidence else slice(None) for a in self.domain]
        newdom = self.domain.marginalize(evidence.keys())
        values = self.values[tuple(slices)]
        return Factor(newdom, values)

    def copy(self, out=None):
        if out is None:
            return Factor(self.domain, self.values.copy())
        np.copyto(out.values, self.values)
        return out

    def __mul__(self, other):
        if np.isscalar(other):
            new_values = np.nan_to_num(other * self.values)
            return Factor(self.domain, new_values)
        # print(self.values.max(), other.values.max(), self.domain, other.domain)
        newdom = self.domain.merge(other.domain)
        factor1 = self.expand(newdom)
        factor2 = other.expand(newdom)
        return Factor(newdom, factor1.values * factor2.values)

    def __add__(self, other):
        if np.isscalar(other):
            return Factor(self.domain, other + self.values)
        newdom = self.domain.merge(other.domain)
        factor1 = self.expand(newdom)
        factor2 = other.expand(newdom)
        return Factor(newdom, factor1.values + factor2.values)

    def __iadd__(self, other):
        if np.isscalar(other):
            self.values += other
            return self
        factor2 = other.expand(self.domain)
        self.values += factor2.values
        return self

    def __imul__(self, other):
        if np.isscalar(other):
            self.values *= other
            return self
        factor2 = other.expand(self.domain)
        self.values *= factor2.values
        return self

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __sub__(self, other):
        if np.isscalar(other):
            return Factor(self.domain, self.values - other)
        other = Factor(
            other.domain, np.where(other.values == -np.inf, 0, -other.values)
        )
        return self + other

    def __truediv__(self, other):
        # assert np.isscalar(other), 'divisor must be a scalar'
        if np.isscalar(other):
            new_values = self.values / other
            new_values = np.nan_to_num(new_values)
            return Factor(self.domain, new_values)
        tmp = other.expand(self.domain)
        vals = np.divide(self.values, tmp.values, where=tmp.values > 0)
        vals[tmp.values <= 0] = 0.0
        return Factor(self.domain, vals)

    def exp(self, out=None):
        if out is None:
            return Factor(self.domain, np.exp(self.values))
        np.exp(self.values, out=out.values)
        return out

    def log(self, out=None):
        if out is None:
            return Factor(self.domain, np.log(self.values + 1e-100))
        np.log(self.values, out=out.values)
        return out

    def datavector(self, flatten=True):
        """Materialize the data vector"""
        if flatten:
            return self.values.flatten()
        return self.values



# A generative model training algorithm based on
# "Winning the NIST Contest: A scalable and general approach to differentially private synthetic data"
# by Ryan McKenna, Gerome Miklau, Daniel Sheldon
# Adapted from: https://github.com/ryan112358/private-pgm/blob/1da21c8b38149b05f1385b8e54116568b700b4fa/mechanisms/mst.py


# import argparse
import itertools
from logging import getLogger
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, Self, Tuple, Optional

import joblib
import networkx as nx
import numpy as np
import pandas as pd
from disjoint_set import DisjointSet
from numpy.random import RandomState
from scipy.special import logsumexp

from dpmm.models.base.graphical import GraphicalGenerativeModel
from dpmm.models.base.mbi import Dataset, Domain, FactoredInference, GraphicalModel
from dpmm.models.base.mechanisms import cdp_rho
from dpmm.models.base.memory import clique_size, model_size
from dpmm.models.base.utils import gaussian_noise
from dpmm.utils import to_path
from dpmm.models.base.mechanisms import Mechanism

"""
This is a generalization of the winning mechanism from the
2018 NIST Differential Privacy Synthetic Data Competition.

Unlike the original implementation, this one can work for any discrete dataset,
and does not rely on public provisional data for measurement selection.
"""


logger = getLogger("dpmm")


def compute_weight(
    est: FactoredInference, data: Dataset, clique: Tuple[str, str]
) -> Tuple[str, str, float, int]:
    """
    Compute the weight of a clique based on the difference between the estimated and actual data vectors.

    :param est: The inference engine used to estimate the data.
    :type est: FactoredInference
    :param data: The dataset.
    :type data: Dataset
    :param clique: A tuple representing the clique (pair of attributes).
    :type clique: Tuple[str, str]
    :return: A tuple containing the clique attributes, weight, and model size.
    :rtype: Tuple[str, str, float, int]
    """
    a, b = clique
    model_size = clique_size(data, (a, b))
    xhat = est.project([a, b]).datavector()
    x = data.project([a, b]).datavector()
    weight = np.linalg.norm(x - xhat, 1)
    return a, b, weight, model_size


class MST(Mechanism):
    """
    Maximum Spanning Tree (MST) mechanism is a differentially private generative model relying
    on selecting an optimal set of marginals to approximate the joint distribution of the data.
    It uses the exponential mechanism to select higher-order marginals based on their weights.
    The marginals are measured using the Laplace mechanism.
    The measured marginals are then used to estimate a maximum spanning tree which will be able to generate data.

    Ref: https://arxiv.org/pdf/2108.04978

    :param epsilon: Privacy budget.
    :type epsilon: float, optional
    :param delta: Privacy parameter.
    :type delta: float, optional
    :param n_iters: Number of iterations for inference.
    :type n_iters: int
    :param compress: Whether to compress the data.
    :type compress: bool
    :param domain: The domain of the data.
    :type domain: Domain, optional
    :param prng: Random state for reproducibility.
    :type prng: RandomState, optional
    :param max_model_size: Maximum model size in MB.
    :type max_model_size: int, optional
    :param structural_zeros: Structural zeros in the data.
    :type structural_zeros: dict, optional
    :param n_jobs: Number of parallel jobs.
    :type n_jobs: int
    """

    def __init__(
        self,
        epsilon: Optional[float] = None,
        delta: Optional[float] = None,
        n_iters: int = 10000,
        compress: bool = True,
        domain: Optional[Domain] = None,
        prng: Optional[RandomState] = None,
        max_model_size: Optional[int] = None,
        structural_zeros: Optional[dict] = None,
        n_jobs: int = -1,
    ):
        super().__init__(
            epsilon=epsilon,
            delta=delta,
            prng=prng,
            max_model_size=max_model_size,
            compress=compress,
            domain=domain,
            structural_zeros=structural_zeros,
            n_jobs=n_jobs,
        )

        self.rho = cdp_rho(self.epsilon, self.delta)
        self.sigma = np.sqrt(3 / (2 * self.rho))
        self.n_iters = n_iters

    def _fit(self, data: Dataset, public: bool = False) -> Tuple[Dataset, list]:
        """
        Fit the MST mechanism to the data.

        :param data: The dataset.
        :type data: Dataset
        :param public: Whether the data is public. Defaults to False.
        :type public: bool, optional
        :return: The dataset and measurement log.
        :rtype: Tuple[Dataset, list]
        """
        # select all 1-way marginals
        self.cliques = cliques_1 = [(col,) for col in data.domain]

        log1 = self.measure(data, cliques=cliques_1, public=public)
        # compress domain of all 1-way marginals
        if self.compress:
            log1 = self.compressor.fit(log1)
            data = self.compressor.transform(data)

        self.model_size = model_size(data, cliques_1)
        # select higher marginals
        if len(data.domain.attrs) > 1:
            cliques_2 = self.select(data, self.rho / 3.0, log1)
            # measure higher marginals
            log2 = self.measure(data, cliques=cliques_2, flatten=True)
            self.cliques += cliques_2
            # TODO document
        else:
            log2 = []

        return data, log1 + log2

    def select(
        self,
        data: Dataset,
        rho: float,
        measurement_log: list,
        cliques: list = [],
        public: bool = False,
    ) -> list:
        """
        Select higher-order marginals using the exponential mechanism.

        :param data: The dataset.
        :type data: Dataset
        :param rho: Privacy budget for selection.
        :type rho: float
        :param measurement_log: Log of measurements.
        :type measurement_log: list
        :param cliques: Existing cliques. Defaults to an empty list.
        :type cliques: list, optional
        :param public: Whether the data is public. Defaults to False.
        :type public: bool, optional
        :return: A list of selected cliques.
        :rtype: list
        """
        engine = FactoredInference(data.domain, iters=2500)
        est = engine.estimate(measurement_log)

        weights = {}
        candidates = list(itertools.combinations(data.domain.attrs, 2))
        if self.n_jobs > 1:
            for a, b, weight, model_size in Pool(self.n_jobs).starmap(
                compute_weight,
                zip(itertools.cycle([est]), itertools.cycle([data]), candidates),
            ):
                weights[a, b] = weight
                self.model_size += model_size
                if self.max_model_size is not None:
                    if self.model_size > self.max_model_size:
                        break
        else:
            for compute_args in zip(itertools.cycle([est]), itertools.cycle([data]), candidates):
                a, b, weight, model_size = compute_weight(*compute_args)
                weights[a, b] = weight
                self.model_size += model_size
                if self.max_model_size is not None:
                    if self.model_size > self.max_model_size:
                        break
                

        T = nx.Graph()
        T.add_nodes_from(data.domain.attrs)
        ds = DisjointSet()

        for e in cliques:
            T.add_edge(*e)
            ds.union(*e)

        r = len(list(nx.connected_components(T)))
        epsilon = np.sqrt(8 * rho / (r - 1))
        for i in range(r - 1):
            candidates = [e for e in candidates if not ds.connected(*e)]
            wgts = np.array([weights[e] for e in candidates])
            if public:
                idx = np.argmax(wgts)
            else:
                idx = self.exponential_mechanism(wgts, epsilon, sensitivity=1.0)

            e = candidates[idx]
            T.add_edge(*e)
            ds.union(*e)

        return list(T.edges)

    def exponential_mechanism(
        self, q: np.ndarray, eps: float, sensitivity: float, monotonic: bool = False
    ) -> int:
        """
        Apply the exponential mechanism to select a candidate.

        :param q: Quality scores for candidates.
        :type q: np.ndarray
        :param eps: Privacy budget.
        :type eps: float
        :param sensitivity: Sensitivity of the scores.
        :type sensitivity: float
        :param monotonic: Whether the scores are monotonic. Defaults to False.
        :type monotonic: bool, optional
        :return: The index of the selected candidate.
        :rtype: int
        """
        coef = 1.0 if monotonic else 0.5
        scores = coef * eps / sensitivity * q
        probas = np.exp(scores - logsumexp(scores))
        return self.prng.choice(q.size, p=probas)

    def store(self, path: Path) -> None:
        """
        Store the MST mechanism state to a file.

        :param path: The path to store the state.
        :type path: Path
        """
        joblib.dump(
            {
                "epsilon": self.epsilon,
                "delta": self.delta,
                "compress": self._compress,
                "compressor": self.compressor,
                "cliques": self.cliques,
                "rho": self.rho,
                "sigma": self.sigma,
                "n_iters": self.n_iters,
                "_domain": self._domain,
                "max_model_size": self.max_model_size,
                "model_size": self.model_size,
                "fit_state": self.fit_state,
            },
            path / "state.joblib",
        )

        if self.model is not None:
            self.model.save(path / "estimator.pickle")

    @classmethod
    def load(cls, path: Path) -> Self:
        """
        Load the MST mechanism state from a file.

        :param path: The path to load the state.
        :type path: Path
        :return: The loaded MST mechanism.
        :rtype: MST
        """
        state = joblib.load(path / "state.joblib")
        obj = cls(
            epsilon=state["epsilon"],
            delta=state["delta"],
            n_iters=state["n_iters"],
            compress=state["compress"],
            domain=state["_domain"],
            max_model_size=state["max_model_size"],
        )
        obj.cliques = state["cliques"]
        obj.fit_state = state["fit_state"]
        obj.compressor = state["compressor"]
        obj.model_size = state["model_size"]
        model_path = path / "estimator.pickle"
        if model_path.exists():
            obj.model = GraphicalModel.load(model_path)
        return obj


class MSTGM(GraphicalGenerativeModel):
    """
    Maximum Spanning Tree (MST) mechanism is a differentially private generative model relying
    on selecting an optimal set of marginals to approximate the joint distribution of the data.
    It uses the exponential mechanism to select higher-order marginals based on their weights.
    The marginals are measured using the Laplace mechanism.
    The measured marginals are then used to estimate a maximum spanning tree which will be able to generate data.

    Ref: https://arxiv.org/pdf/2108.04978

    :param epsilon: Privacy budget.
    :type epsilon: float
    :param delta: Privacy parameter.
    :type delta: float
    :param n_iters: Number of iterations for inference.
    :type n_iters: int
    :param compress: Whether to compress the data.
    :type compress: bool
    :param max_model_size: Maximum model size in MB.
    :type max_model_size: int, optional
    :param domain: The domain of the data.
    :type domain: Domain, optional
    :param random_state: Random state for reproducibility.
    :type random_state: RandomState, optional
    """

    name = "mst"

    def __init__(
        self,
        epsilon: float = 1,
        delta: float = 1e-5,
        n_iters: int = 5000,
        compress: bool = True,
        max_model_size: Optional[int] = None,
        domain: Optional[Domain] = None,
        random_state: Optional[RandomState] = None,
        n_jobs: int = -1,
    ):
        super().__init__(domain=domain, random_state=random_state)

        self.epsilon = epsilon
        self.delta = delta

        self.generator = MST(
            epsilon=self.epsilon,
            delta=self.delta,
            n_iters=n_iters,
            compress=compress,
            domain=domain,
            prng=random_state,
            max_model_size=max_model_size,
            n_jobs=n_jobs,
        )

    def set_random_state(self, random_state: RandomState) -> None:
        """
        Set the random state for reproducibility.

        :param random_state: The random state.
        :type random_state: RandomState
        """
        super().set_random_state(random_state)
        self.generator.set_random_state(random_state)

    def set_domain(self, domain: Dict) -> None:
        """
        Set the domain of the data.

        :param domain: The domain.
        :type domain: Dict
        """
        super().set_domain(domain=domain)
        self.generator.set_domain(domain=domain)

    @to_path
    def store(self, path: Path) -> None:
        """
        Store the MSTGM model state to a file.

        :param path: The path to store the state.
        :type path: Path
        """
        super().store(path)
        self.generator.store(path)

    @classmethod
    @to_path
    def load(cls, path: Path) -> Self:
        """
        Load the MSTGM model state from a file.

        :param path: The path to load the state.
        :type path: Path
        :return: The loaded MSTGM model.
        :rtype: MSTGM
        """
        generator = MST.load(path)
        obj = cls(epsilon=generator.epsilon, delta=generator.delta)
        del obj.generator
        obj.generator = generator

        return obj



from dpmm.models.base.mbi.domain import Domain
from dpmm.models.base.mbi.dataset import Dataset
from dpmm.models.base.mbi.callbacks import Logger
from dpmm.models.base.mbi.factor import Factor
from dpmm.models.base.mbi.clique_vector import CliqueVector
from dpmm.models.base.mbi.junction_tree import JunctionTree
from dpmm.models.base.mbi.graphical_model import GraphicalModel
from dpmm.models.base.mbi.inference import FactoredInference



import numpy as np
from opendp.measurements import make_laplace, make_gaussian
from opendp.mod import enable_features
import opendp.prelude as dp


def gaussian_noise(sigma, size=None):
    enable_features("floating-point", "contrib")
    input_domain = dp.atom_domain(T=float)
    input_metric = dp.absolute_distance(T=float)
    meas = make_gaussian(input_domain, input_metric, sigma)
    if size is None:
        return meas(0.0)
    elif isinstance(size, np.number):
        return np.array([meas(0.0) for _ in range(size)])
    else:
        _size = np.prod(size)
        measurements = [meas(0.0) for _ in range(_size)]
        return np.array(measurements).reshape(size)


def laplace_noise(scale, size=None):
    enable_features("floating-point", "contrib")
    input_domain = dp.atom_domain(T=float)
    input_metric = dp.absolute_distance(T=float)
    meas = make_laplace(input_domain, input_metric, scale)
    if size is None:
        return meas(0.0)
    else:
        return [meas(0.0) for _ in range(size)]



from pathlib import Path
from typing import Callable, Any, Union


def to_path(func: Callable) -> Callable:
    """
    Decorator to convert a string path argument to a `Path` object.

    :param func: The function to decorate.
    :return: The decorated function.

    .. code-block:: python

        @to_path
        def save_file(self, path: Path):
            path.write_text("Hello, World!")
    """
    def new_func(self: Any, path: Union[str, Path]) -> Any:
        return func(self, Path(path))

    return new_func



import numpy as np
import pandas as pd
from typing import List, Tuple, Optional, Dict, Union
from numpy.random import RandomState

from dpmm.models.base.utils import laplace_noise
from dpmm.processing.utils import approx_bounds, optimal_n_bins

cast_map = {
    "f": np.float32,
    "i": np.int32,
    "u": np.uint32,
    "b": np.bool_,
    "M": np.datetime64,
    "m": lambda x: pd.Timedelta(x).to_numpy(),
}


def get_cast(dtype: np.dtype) -> Union[np.dtype, callable]:
    """
    Get the appropriate casting function or dtype for a given numpy dtype.

    :param dtype: The numpy dtype to map.
    :return: The corresponding casting function or dtype.

    .. code-block:: python

        dtype = np.dtype('float32')
        cast_func = get_cast(dtype)
    """
    return cast_map[dtype.kind]


class Binner:
    name: str = None

    def __init__(
        self,
        n_bins: Union[str, int] = "auto",
        lower: Optional[float] = None,
        upper: Optional[float] = None,
        epsilon: Optional[float] = None,
        max_n_bins: Optional[int] = None,
        rnd: Optional[RandomState] = None,
        structural_zeros: Optional[List[Tuple[float, float]]] = None,
    ):
        """
        Initialize the Binner.

        :param n_bins: Number of bins or 'auto' for automatic binning.
        :param lower: Lower bound of the data.
        :param upper: Upper bound of the data.
        :param epsilon: Privacy budget.
        :param max_n_bins: Maximum number of bins.
        :param rnd: Random state for reproducibility.
        :param structural_zeros: List of intervals treated as structural zeros.

        .. code-block:: python

            binner = Binner(n_bins=10, lower=0, upper=1, epsilon=0.1)
        """
        self.n_bins = n_bins
        self.lower = lower
        self.upper = upper
        self.epsilon = epsilon
        self.spent_epsilon = None
        self.max_n_bins = max_n_bins
        self.dtype = None
        if structural_zeros is None:
            structural_zeros = []

        self.set_structural_zeros(structural_zeros)

        # Random State
        if rnd is None:
            rnd = RandomState()
        self.rnd = rnd

    def set_structural_zeros(self, structural_zeros: List[Tuple[float, float]]) -> None:
        """
        Set structural zeros for the binner.

        :param structural_zeros: A list of tuples representing intervals to be treated as structural zeros.

        .. code-block:: python

            binner.set_structural_zeros([(0.1, 0.2), (0.5, 0.6)])
        """
        structural_zeros = sorted(structural_zeros)
        # remove overlapping zeros
        if len(structural_zeros) >= 1:
            new_zeros = [structural_zeros[0]]
            for zero in structural_zeros[1:]:
                # if previous interval is not overlapping with the current interval
                if zero[0] > new_zeros[-1][1]:
                    new_zeros.append(zero)
                elif zero[1] > new_zeros[-1][1]:
                    new_zeros[-1] = (new_zeros[-1][0], zero[1])

            structural_zeros = new_zeros

        self.structural_zeros = structural_zeros

    def set_random_state(self, rnd: RandomState) -> None:
        """
        Set the random state for reproducibility.

        :param rnd: A numpy RandomState instance.

        .. code-block:: python

            rnd = RandomState(42)
            binner.set_random_state(rnd)
        """
        self.rnd = rnd

    def spend_epsilon(self, eps: float) -> float:
        """
        Spend a portion of the privacy budget (epsilon).

        :param eps: The amount of privacy budget (epsilon) to spend.
        :return: The spent privacy budget.

        .. code-block:: python

            spent_eps = binner.spend_epsilon(0.1)
        """
        if self.epsilon is None:
            raise ValueError("Should not spend any budget as the overall epsilon")
        else:
            if self.spent_epsilon is None:
                self.spent_epsilon = eps
            else:
                self.spent_epsilon += eps
        return eps

    @property
    def zeros(self) -> np.ndarray:
        """
        Get the indices of bins that are structural zeros.

        :return: An array of indices of structural zero bins.

        .. code-block:: python

            zero_indices = binner.zeros
        """
        col_zeros = np.array(self.structural_zeros)  # n_zeroes x 2 (start, end)
        bins = np.array([self.bins[:-1], self.bins[1:]]).T  # n_bins x 2 (start, end)
        bin_is_in_interval = np.logical_and(
            col_zeros[np.newaxis, :, 0] <= bins[:, np.newaxis, 0],
            col_zeros[np.newaxis, :, 1] >= bins[:, np.newaxis, 1],
        )  # n_bins x n_zeroes
        is_zero = np.any(bin_is_in_interval, axis=1)
        return np.arange(len(is_zero))[is_zero]

    def get_bounds(self, arr: np.ndarray, recompute: bool = False, epsilon: Optional[float] = None) -> Tuple[float, float]:
        """
        Compute or retrieve the bounds of the data.

        :param arr: The input array.
        :param recompute: Whether to recompute the bounds.
        :param epsilon: The privacy budget for computing bounds.
        :return: A tuple containing the lower and upper bounds.

        .. code-block:: python

            bounds = binner.get_bounds(arr, epsilon=0.1)
        """
        
        if self.has_bounds and not recompute:
            min_, max_ = self.lower, self.upper
        elif epsilon is None:
            min_, max_ = arr.min(), arr.max()
            if pd.notnull(self.lower):
                min_ = self.lower
            if pd.notnull(self.upper):
                max_ = self.upper
        else:
            min_, max_ = approx_bounds(arr, epsilon=self.spend_epsilon(epsilon))

        type_cast = get_cast(self.dtype)
        min_, max_ = type_cast(min_), type_cast(max_)
        return min_, max_

    def bin_bounds(self, bin_idx: int) -> Tuple[float, float]:
        """
        Get the bounds of a specific bin.

        :param bin_idx: The index of the bin.
        :return: A tuple containing the lower and upper bounds of the bin.

        .. code-block:: python

            bin_bounds = binner.bin_bounds(3)
        """
        return self.bins[bin_idx], self.bins[bin_idx + 1]

    @property
    def has_bounds(self) -> bool:
        """
        Check if the binner has predefined bounds.

        :return: True if bounds are defined, False otherwise.

        .. code-block:: python

            has_bounds = binner.has_bounds
        """
        return (self.lower is not None) and (self.upper is not None)

    def fit(self, arr: np.ndarray) -> None:
        """
        Fit the binner to the data.

        :param arr: The input array to fit.

        .. code-block:: python

            binner.fit(arr)
        """
        if isinstance(self.n_bins, str):
            self.n_bins = optimal_n_bins(
                arr, upper=self.max_n_bins, epsilon=self.epsilon
            )
        self.dtype = arr.dtype

    def transform(self, arr: np.ndarray) -> np.ndarray:
        """
        Transform the data into bins.

        :param arr: The input array to transform.
        :return: The transformed array.

        .. code-block:: python

            transformed = binner.transform(arr)
        """
        raise NotImplementedError

    def fit_transform(self, arr: np.ndarray) -> np.ndarray:
        """
        Fit the binner to the data and transform it.

        :param arr: The input array to fit and transform.
        :return: The transformed array.

        .. code-block:: python

            transformed = binner.fit_transform(arr)
        """
        self.fit(arr)
        return self.transform(arr)

    def inverse_transform(self, arr: np.ndarray) -> np.ndarray:
        """
        Inverse transform the binned data back to its original form.

        :param arr: The binned array to inverse transform.
        :return: The original array.

        .. code-block:: python

            original = binner.inverse_transform(binned_arr)
        """
        raise NotImplementedError

    @property
    def bin_domain(self) -> int:
        """
        Get the number of bins minus one.

        :return: The number of bins minus one.

        .. code-block:: python

            domain = binner.bin_domain
        """
        return len(self.bins) - 1


class UniformBinner(Binner):
    name: str = "uniform"

    def __init__(
        self,
        n_bins: int = 20,
        epsilon: Optional[float] = None,
        lower: Optional[float] = None,
        upper: Optional[float] = None,
        max_n_bins: Optional[int] = None,
        rnd: Optional[RandomState] = None,
    ):
        """
        Initialize the UniformBinner. Distribute the data into uniform bins.
        The number of bins is determined by the `n_bins` parameter.

        :param n_bins: Number of bins.
        :param epsilon: Privacy budget.
        :param lower: Lower bound of the data.
        :param upper: Upper bound of the data.
        :param max_n_bins: Maximum number of bins.
        :param rnd: Random state for reproducibility.

        .. code-block:: python

            binner = UniformBinner(n_bins=10)
        """
        super().__init__(
            n_bins=n_bins,
            lower=lower,
            upper=upper,
            epsilon=epsilon,
            max_n_bins=max_n_bins,
            rnd=rnd,
        )
        self.bins = None
        self.stats = {}

    def fit(self, arr: np.ndarray) -> None:
        """
        Fit the UniformBinner to the data.

        :param arr: The input array to fit.

        .. code-block:: python

            binner.fit(arr)
        """
        super().fit(arr)

        if self.epsilon is None:
            eps = None
        else:
            eps = self.epsilon

        lower_, upper_ = self.get_bounds(arr=arr, epsilon=eps)
        self.bins = lower_ + (upper_ - lower_) * np.linspace(0, 1, num=self.n_bins + 1)

    def transform(self, arr: np.ndarray) -> np.ndarray:
        """
        Transform the data into bins.

        :param arr: The input array to transform.
        :return: The transformed array.

        .. code-block:: python

            transformed = binner.transform(arr)
        """

        bins = np.array(self.bins).astype(arr.dtype)
        labels = [
            idx for idx, bin in enumerate(self.bins[:-1]) if self.bins[idx + 1] != bin
        ]

        t_series = (
            pd.Series(
                pd.cut(
                    arr.clip(bins[0], bins[-1]),
                    bins=self.bins,
                    right=False,
                    labels=labels,
                    duplicates="drop",
                )
            )
            .fillna(labels[-1])
            .astype(int)
            .clip(lower=0, upper=(len(self.bins) - 2))
        )

        return t_series.to_numpy()

    def get_intervals(self, bin_idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get the intervals and weights for a specific bin.

        :param bin_idx: The index of the bin.
        :return: A tuple containing the intervals and their weights.

        .. code-block:: python

            intervals, weights = binner.get_intervals(3)
        """
        lower, upper = self.bins[bin_idx], self.bins[bin_idx + 1]
        intervals = np.array([(lower, upper)])

        if len(self.structural_zeros) > 0:
            left_zeros = np.array([zero[0] for zero in self.structural_zeros])
            right_zeros = np.array([zero[1] for zero in self.structural_zeros])

            # if lower is in an interval that stops before upper bound
            lower_in_zero = (
                (lower >= left_zeros) & (lower <= right_zeros) & (right_zeros <= upper)
            )
            if lower_in_zero.any():
                lower = right_zeros[lower_in_zero].max()

            # update upper
            upper_in_zero = (
                (lower <= left_zeros) & (upper >= left_zeros) & (upper <= right_zeros)
            )
            if upper_in_zero.any():
                upper = left_zeros[upper_in_zero].min()

            # Create new intervals to sample from
            _steps = (
                [lower]
                + [
                    z
                    for zero in self.structural_zeros
                    for z in zero
                    if (zero[0] >= lower) and (zero[1] <= upper)
                ]
                + [upper]
            )
            intervals = np.array(
                [
                    (step, _steps[idx + 1])
                    for idx, step in list(enumerate(_steps[:-1]))[::2]
                ]
            )

        weights = np.array([z[1] - z[0] for z in intervals])
        weights = weights / weights.sum()
        return intervals, weights

    def inverse_transform(self, arr: np.ndarray) -> np.ndarray:
        """
        Inverse transform the binned data back to its original form.

        :param arr: The binned array to inverse transform.
        :return: The original array.

        .. code-block:: python

            original = binner.inverse_transform(binned_arr)
        """
        i_series = pd.Series(index=np.arange(arr.shape[0]), dtype=object)
        for bin_idx in range(self.n_bins):
            intervals, weights = self.get_intervals(bin_idx)

            bin_match = arr == bin_idx
            int_idx = self.rnd.choice(len(intervals), size=bin_match.sum(), p=weights)
            new_series = (
                self.rnd.uniform(0, 1, size=bin_match.sum())
                * (intervals[int_idx, 1] - intervals[int_idx, 0])
                + intervals[int_idx, 0]
            )
            # if "timedelta" in self.dtype.name:
            i_series.loc[bin_match] = new_series


        if self.dtype.kind in "ui":
            i_series = i_series.round().astype(self.dtype)
        return i_series.to_numpy()


class PrivTreeBinner(UniformBinner):
    name: str = "priv-tree"
    beta = 2

    def __init__(
        self,
        n_bins: int = 20,
        enforce_bins: bool = True,
        epsilon: Optional[float] = None,
        lower: Optional[float] = None,
        upper: Optional[float] = None,
        max_n_bins: Optional[int] = None,
        decay: float = 0,
        rnd: Optional[RandomState] = None,
    ):
        """
        Initialize the PrivTreeBinner. 
        Based on the (PrivTree Algorithm)[https://arxiv.org/pdf/1601.03229].
        

        :param n_bins: Number of bins.
        :param enforce_bins: Whether to enforce the number of bins.
        :param epsilon: Privacy budget.
        :param lower: Lower bound of the data.
        :param upper: Upper bound of the data.
        :param max_n_bins: Maximum number of bins.
        :param decay: Decay factor for the privacy budget.
        :param rnd: Random state for reproducibility.

        .. code-block:: python

            binner = PrivTreeBinner(n_bins=10, enforce_bins=True, epsilon=0.1)
        """
        super().__init__(
            n_bins=n_bins,
            lower=lower,
            epsilon=epsilon,
            upper=upper,
            max_n_bins=max_n_bins,
            rnd=rnd,
        )
        self.bins = None
        self.enforce_bins = enforce_bins
        self.theta = None
        self.decay = decay
        self.scale = None

    def priv_tree(self, data: np.ndarray, eps: Optional[float] = None) -> Tuple[Dict, Dict, Dict, Dict, List[str]]:
        """
        Builds bin by splitting the data into a binary tree structure.
        The tree is built by recursively splitting the data into two halves
        until the number of data points in a node is less than theta.
        The tree is built in a top-down manner, where each node represents
        a sub-domain of the data.
        The tree is built using the Laplace mechanism to ensure differential privacy.
        The tree is built in a way that the number of bins is equal to n_bins.

        :param data: The input data array.
        :param eps: The privacy budget for the tree.
        :return: A tuple containing domains, children, counts, depths, and leaf nodes.

        .. code-block:: python

            domains, children, counts, depths, leaf_nodes = binner.priv_tree(data, eps=0.1)
        """
        if len(data.shape) == 1:
            lower_, upper_ = self.get_bounds(arr=data, epsilon=eps)
            data = data[:, np.newaxis]
            lower_, upper_ = np.array([lower_]), np.array([upper_])
        else:
            # TODO: Add get_bounds in multi dimensions
            lower_, upper_ = data.min(axis=0), data.max(axis=0)

        domains = {"v_0": (lower_, upper_, data)}
        children = {}
        counts = {"v_0": data.shape[0]}
        depths = {"v_0": 0}
        node_idx = 1
        unvisited = ["v_0"]
        leaf_nodes = []

        while unvisited:
            # get unvisited node
            node = unvisited[0]  # node name
            node_count = counts[node]  # node sub-domain count
            node_depth = depths[node]  # node depth
            node_domain = domains[node]

            # remove node from unvisited
            unvisited = unvisited[1:]

            # biased count
            b_node = max(
                (node_count - (node_depth * self.decay)), (self.theta - self.decay)
            )

            # add laplace noise
            if self.scale is not None:
                b_node += laplace_noise(scale=self.scale)

            if (b_node > self.theta) and (
                (not self.enforce_bins)
                or ((len(leaf_nodes) + len(unvisited) + 2) <= self.n_bins)
            ):
                node_split = node_domain[0] + (node_domain[1] - node_domain[0]) / 2
                children[node] = []

                # Split domain
                child_domains = [node_domain]
                for axis_idx in range(node_split.shape[0]):
                    new_domains = []
                    for domain_lower, domain_upper, domain_data in child_domains:
                        node_split = domain_lower + (domain_upper - domain_lower) / 2
                        split_flag = domain_data[:, axis_idx] < node_split[axis_idx]
                        left = domain_data[split_flag]
                        right = domain_data[~split_flag]
                        new_domains += [
                            (domain_lower, node_split, left),
                            (node_split, domain_upper, right),
                        ]

                    child_domains = new_domains

                # Update list
                for child_node in child_domains:
                    node_name = f"v_{node_idx}"
                    # Update children
                    children[node].append(node_name)
                    # Update domains
                    domains[node_name] = child_node
                    # Update counts
                    counts[node_name] = child_node[-1].shape[0]
                    # Update depth
                    depths[node_name] = node_depth + 1
                    # Update unvisited
                    unvisited.append(node_name)
                    # Update node index
                    node_idx += 1
            else:
                leaf_nodes.append(node)

        return domains, children, counts, depths, leaf_nodes

    def fit(self, arr: np.ndarray) -> None:
        """
        Fit the PrivTreeBinner to the data.

        :param arr: The input array to fit.

        .. code-block:: python

            binner.fit(arr)
        """
        Binner.fit(self, arr)
        self.theta = arr.shape[0] / self.n_bins
        eps = self.epsilon
        if self.epsilon:
            eps = self.epsilon / (1 + int(not (self.has_bounds)))

            self.scale = (2 * self.beta - 1) / (
                (self.beta - 1) * self.spend_epsilon(eps)
            )  # ln(2) to be changed to ln(2^dim)
            self.decay = self.scale * np.log(2)  # ln(2) to be changed to ln(2^dim)

        (
            self.domains,
            self.children,
            self.counts,
            self.depths,
            self.leaf_nodes,
        ) = self.priv_tree(data=arr, eps=eps)

        sorted_leaves = sorted(self.leaf_nodes, key=lambda x: tuple(self.domains[x][0]))
        self.bins = [self.domains[node][0][0] for node in sorted_leaves] + [
            self.domains[sorted_leaves[-1]][1][0]
        ]

        self.n_bins = len(self.bins) - 1
